// Fill out your copyright notice in the Description page of Project Settings.

#include "LoadTest.h"
#include "AssetStreamer.h"

#include "IPlatformFilePak.h"
#include "StreamingNetworkPlatformFile.h"
#include "FileManagerGeneric.h"

FAssetStreamer::FAssetStreamer():
	PakPlatform(new FPakPlatformFile())
	, bInitialized(false)
	, StreamableManager(nullptr)
{

}

bool FAssetStreamer::Initialize(FStreamableManager* StreamableManager)
{
	check(StreamableManager);
	if (!bInitialized)
	{

		LocalPlatformFile = &FPlatformFileManager::Get().GetPlatformFile();
		if (LocalPlatformFile != nullptr)
		{
			IPlatformFile* PakPlatformFile = FPlatformFileManager::Get().GetPlatformFile(TEXT("PakFile"));
		}
		bInitialized = true;
	}
	this->StreamableManager = StreamableManager;
	return bInitialized;
}

void FindAllPakFiles(IPlatformFile* LowLevelFile, const TArray<FString> Directory, TArray<FString>& OutPakFiles)
{
	// Helper class to find all pak files.
	class FPakSearchVisitor : public IPlatformFile::FDirectoryVisitor
	{
		TArray<FString>& FoundPakFiles;
		IPlatformChunkInstall* ChunkInstall;
	public:
		FPakSearchVisitor(TArray<FString>& InFoundPakFiles, IPlatformChunkInstall* InChunkInstall)
			: FoundPakFiles(InFoundPakFiles)
			, ChunkInstall(InChunkInstall)
		{}
		virtual bool Visit(const TCHAR* FilenameOrDirectory, bool bIsDirectory)
		{
			if (bIsDirectory == false)
			{
				FString Filename(FilenameOrDirectory);
				if (FPaths::GetExtension(Filename) == TEXT("pak"))
				{
					FoundPakFiles.Add(Filename);
				}
			}
			return true;
		}
	};
	// Find all pak files.
	FPakSearchVisitor Visitor(OutPakFiles, FPlatformMisc::GetPlatformChunkInstall());
	for (auto Iter(Directory.CreateConstIterator()); Iter; Iter++)
	{
		LowLevelFile->IterateDirectoryRecursively(**Iter, Visitor);
	}
}

bool FAssetStreamer::StreamPackage(const FString& PakFileName, IAssetStreamerListener* AssetStreamerListener)
{
	Lock();
	Listener = NULL;

	if (!UseLocal())
	{
		Unlock();
		return false;
	}
	FPlatformFileManager::Get().SetPlatformFile(*PakPlatform);

	// Now Get the path and start the streaming
	const FString FilePath = PakFileName + TEXT(".pak");
	TArray<FString> PakFolders;
	FPakPlatformFile::GetPakFolders(TEXT(""), PakFolders);
	TArray<FString> FoundPakFiles;
	FindAllPakFiles(LocalPlatformFile, PakFolders, FoundPakFiles);
	for (auto Iter(FoundPakFiles.CreateConstIterator()); Iter; Iter++)
	{
		UE_LOG(LogClass, Log, TEXT("**File %s "), **Iter);
	}
	if (!PakPlatform->FileExists(*FilePath))
	{
		UE_LOG(LogClass, Error, TEXT("File %s do not exist Try another one "), *FilePath);
		return false;
	}
	// Make sure the Pak file is actually there
	FPakFile PakFile(*FilePath, false);
	if (!PakFile.IsValid())
	{
		Unlock();
		UE_LOG(LogClass, Error, TEXT("Invalid pak file: %s"), *FilePath);
		return false;
	}
	
	PakFile.SetMountPoint(*FPaths::GameContentDir());
	const int32 PakOrder = 0;
	if (!PakPlatform->Mount(*FilePath, PakOrder, *FPaths::GameContentDir()))
	{
		Unlock();
		UE_LOG(LogClass, Error, TEXT("Failed to mount pak file: %s"), *FilePath);
		return false;
	}

	// Load all assets contained in this Pak file
	TSet<FString> FileList;
	PakFile.FindFilesAtPath(FileList, *PakFile.GetMountPoint(), true, false, true);
	// Discover assets within the PakFile
	StreamedAssets.Empty();
	for (TSet<FString>::TConstIterator FileItem(FileList); FileItem; ++FileItem)
	{
		FString AssetName = *FileItem;
		if (AssetName.EndsWith(FPackageName::GetAssetPackageExtension()) ||
			AssetName.EndsWith(FPackageName::GetMapPackageExtension()))
		{
			if (AssetName.RemoveFromStart("../../../LoadTest/Content"))
			{
				//if (AssetName.Contains("ArchVis") && AssetName.RemoveFromEnd(".uasset"))
				//{
					StreamedAssets.Add("/Game" + AssetName);
				//}
			}
		}
	}

	// Once we started the async work assign listener
	Listener = AssetStreamerListener;

	// Tell the listener which assets we are about to stream
	if (Listener)
	{
		Listener->OnPrepareAssetStreaming(StreamedAssets);
	}
	StreamableManager->RequestAsyncLoad(StreamedAssets, FStreamableDelegate::CreateRaw(this, &FAssetStreamer::OnStreamingCompleteDelegate));
	return true;
}

bool FAssetStreamer::UseLocal()
{
	if (PakPlatform.IsValid())
	{
		IPlatformFile* InnerPlatform = LocalPlatformFile;
		return InnerPlatform && PakPlatform->Initialize(InnerPlatform, TEXT(""));
	}
	UE_LOG(LogClass, Error, TEXT("Failed to initialize local file system! ou wont be able to load PAK files!"));
	return false;
}

void FAssetStreamer::OnStreamingCompleteDelegate()
{
	// Call listener!
	if (Listener != NULL)
	{
		Listener->OnAssetStreamComplete();
		Listener = NULL;
	}

	// Finally unlock the streamer
	Unlock();
}