// Fill out your copyright notice in the Description page of Project Settings.

#pragma once


/**
* Interface
*/
class LOADTEST_API IAssetStreamerListener
{
public:
	/** Must have the same signature as FStreamableDelegate. Called when the streaming has completed. */
	virtual void OnAssetStreamComplete() = 0;

	/** Called before the streaming is initiated, used to inform which assets will be requested */
	virtual void OnPrepareAssetStreaming(const TArray<FStringAssetReference>& StreamedAssets) = 0;
};

/**
* The asset streamer is capable of loading and mounting the content of a PAK file from the local FS
*/
class LOADTEST_API FAssetStreamer
{
public:
	FAssetStreamer();

	/** Unlock the streamer when destroyed */
	~FAssetStreamer()
	{
		Unlock();
	}

	/** Initialize the FAssetStreamer and point to the using host:port */
	bool Initialize(FStreamableManager* StreamableManager);

	/**
	* Stream assets form a PAK file remotely or from the local file system
	*/
	bool StreamPackage(const FString& PakFileName, IAssetStreamerListener* AssetStreamerListener);

	/** Check if the streamed has been initialized or not */
	bool bInitialized;
	
	bool IsStreaming() const
	{
		return LockValue != 0;
	}

	/** Blocks this thread until we have stopped streaming */
	void BlockUntilStreamingFinished(float SleepTime = 0.1f) const
	{
		while (LockValue != 0)
		{
			FPlatformProcess::Sleep(SleepTime);
		}
	}

	/** The host:port connection string for the remote file server*/
	FString ServerHost;

	TArray<FStringAssetReference> GetStreamedAsset()
	{
		return StreamedAssets;
	}

private:

	/** Set the  streaming to be using the local FS*/
	bool UseLocal();

	/** Locks the primitive so no one can proceed past a block point */
	void Lock()
	{
		FPlatformAtomics::InterlockedExchange(&LockValue, 1);
	}

	/** Unlocks the primitive so other threads can proceed past a block point */
	void Unlock()
	{
		FPlatformAtomics::InterlockedExchange(&LockValue, 0);
	}

	/** Called once streaming has finished, will call our listener and unlock the streamer again */
	void OnStreamingCompleteDelegate();

	/** Thread safe lock value */
	volatile int32 LockValue;

	// The platform file instances that will help us perform local and remote loading
	IPlatformFile* LocalPlatformFile;
	TSharedPtr<class FPakPlatformFile> PakPlatform;

	/** The current listener for the current streaming task */
	IAssetStreamerListener* Listener;

	/** The assets that we are goind to stream */
	TArray<FStringAssetReference> StreamedAssets;


	/** The streamable manager 'someone' must provide us with from the outside */
	FStreamableManager* StreamableManager;

};